"""
Service base specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec, SpecParams


class ServiceSpec(Spec):
    """
    Service specifications.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """


class ServiceParams(SpecParams):
    """
    Service parameters.
    """



"""
Runtime factory module.
"""
from __future__ import annotations

import importlib
import typing

if typing.TYPE_CHECKING:
    from digitalhub_core.runtimes.base import Runtime
    from digitalhub_core.runtimes.registry import RuntimeRegistry


def build_runtime(kind: str) -> Runtime:
    """
    Runtime factory.

    Returns a runtime instance for the given kind. It looks for the
    runtime registry, imports the specific runtime module, and finally
    get the class from the module by its name.

    Parameters
    ----------
    kind : str
        The runtime kind.

    Returns
    -------
    Runtime
        Runtime instance.
    """
    try:
        registry = import_registry(kind)
        module = importlib.import_module(registry.module)
        return getattr(module, registry.class_name)()
    except (ModuleNotFoundError, ImportError):
        raise ValueError(f"Runtime {kind} not found")


def import_registry(kind: str) -> RuntimeRegistry:
    """
    Import registry from implemented module.

    Parameters
    ----------
    kind : str
        Module kind. Basically, the kind of the function due to correspondence
        function-runtime name.

    Returns
    -------
    dict
        Registry.
    """

    # Cycle over digitalhub layers modules (data, ml, ai).
    for layer in ["core", "data", "ml", "ai"]:
        # Try to import module
        module_name = f"digitalhub_{layer}_{kind}"
        try:
            # Check if module is already imported in cache, otherwise import it
            if module_name not in _modules_cache:
                _modules_cache[module_name] = importlib.import_module(module_name)

            module = _modules_cache[module_name]
            # If module is imported succesfully, break, otherwise continue
            break

        except ModuleNotFoundError:
            continue
    # If module is not imported, raise error
    else:
        raise ModuleNotFoundError(f"Module not found in digitalhub layers data, ml, ai for runtime {kind}")

    # Get registry with classes string pointers
    try:
        return getattr(module, "registry")
    except AttributeError:
        raise ValueError(f"Registry not found in module {module_name}")


_modules_cache = {}



"""
Models for task specifications.
"""
from __future__ import annotations

from typing import Union

from pydantic import BaseModel
from typing_extensions import Literal


class Item(BaseModel):
    """
    Item model.
    """

    key: str
    """Item name."""

    path: str
    """Item mount path inside the container."""


class ConfigMap(BaseModel):
    """
    Spec for config map volume.
    """

    name: str
    """Config map name."""

    items: list[Item] = None
    """List of keys to mount into the container in the specified path.
    If None, all items will be mounted."""


class Secret(BaseModel):
    """
    Spec for secret volume.
    """

    secret_name: str
    """Secret name."""

    items: list[Item] = None
    """List of keys to mount into the container in the specified path.
    If None, all items will be mounted."""


class PVC(BaseModel):
    """
    Spec for persistent volume claim volume.
    """

    claim_name: str
    """Persistent volume claim name."""


class Volume(BaseModel):
    """
    Volume model.
    """

    volume_type: Literal["config_map", "secret", "persistent_volume_claim"]
    """Volume type."""

    name: str
    """Volume name."""

    mount_path: str
    """Volume mount path inside the container."""

    spec: Union[ConfigMap, Secret, PVC]


class NodeSelector(BaseModel):
    """
    NodeSelector model.
    """

    key: str
    """Node selector key."""

    value: str
    """Node selector value."""


class Resource(BaseModel):
    """
    Resource model.
    """

    resource_type: Literal["cpu", "memory", "gpu"]
    """Resource kind (cpu, memory, gpu)."""

    requests: str
    """Resource requests."""

    limits: str
    """Resource limits."""


class Env(BaseModel):
    """
    Env variable model.
    """

    name: str
    """Env variable name."""

    value: str
    """Env variable value."""


class Label(BaseModel):
    """
    Label model.
    """

    name: str
    """Label name."""

    value: str
    """Label value."""


### Affinity and Tolerations ###


class Toleration(BaseModel):
    """
    Toleration model.
    """

    key: str = None
    """Toleration key."""

    operator: str = None
    """Toleration operator."""

    value: str = None
    """Toleration value."""

    effect: str = None
    """Toleration effect."""

    toleration_seconds: int = None
    """Toleration seconds."""


class V1NodeSelectorRequirement(BaseModel):
    key: str
    operator: str
    values: list[str] = None


class V1NodeSelectorTerm(BaseModel):
    match_expressions: list[V1NodeSelectorRequirement] = None
    match_fields: list[V1NodeSelectorRequirement] = None


class V1NodeSelector(BaseModel):
    node_selector_terms: list[V1NodeSelectorTerm]


class V1PreferredSchedulingTerm(BaseModel):
    preference: V1NodeSelector
    weight: int


class V1LabelSelectorRequirement(BaseModel):
    key: str
    operator: str
    values: list[str] = None


class V1LabelSelector(BaseModel):
    match_expressions: list[V1LabelSelectorRequirement] = None
    match_labels: dict[str, str] = None


class V1PodAffinityTerm(BaseModel):
    label_selector: V1LabelSelector = None
    match_label_keys: list[str] = None
    mismatch_label_keys: list[str] = None
    namespace_selector: V1LabelSelector = None
    namespaces: list[str] = None
    topology_key: str


class V1WeightedPodAffinityTerm(BaseModel):
    pod_affinity_term: V1PodAffinityTerm
    weight: int


class V1NodeAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: list[V1PreferredSchedulingTerm] = None
    required_during_scheduling_ignored_during_execution: V1NodeSelector = None


class V1PodAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: list[V1WeightedPodAffinityTerm] = None
    required_during_scheduling_ignored_during_execution: list[V1PodAffinityTerm] = None


class V1PodAntiAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: list[V1WeightedPodAffinityTerm] = None
    required_during_scheduling_ignored_during_execution: list[V1PodAffinityTerm] = None


class Affinity(BaseModel):
    """
    Affinity model.
    """

    node_affinity: V1NodeAffinity = None
    """Node affinity."""

    pod_affinity: V1PodAffinity = None
    """Pod affinity."""

    pod_anti_affinity: V1PodAntiAffinity = None
    """Pod anti affinity."""



"""
Project operations module.
"""
from __future__ import annotations

import importlib.util as imputil
import typing
from pathlib import Path

from digitalhub_core.client.builder import build_client, get_client
from digitalhub_core.context.builder import delete_context
from digitalhub_core.entities.projects.entity import project_from_dict, project_from_parameters
from digitalhub_core.utils.api import api_base_delete, api_base_read, api_base_update
from digitalhub_core.utils.exceptions import BackendError, EntityError
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.projects.entity import Project


ENTITY_TYPE = "projects"


def create_project(**kwargs) -> Project:
    """
    Create a new project.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        A Project instance.
    """
    return project_from_parameters(**kwargs)


def create_project_from_dict(obj: dict) -> Project:
    """
    Create a new Project instance from a dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create object from.

    Returns
    -------
    Project
        Project object.
    """
    return project_from_dict(obj)


def load_project(
    name: str | None = None,
    filename: str | None = None,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Load project and context from backend or file.

    Parameters
    ----------
    name : str
        Project name.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore env configuration.
    setup_kwargs : dict
        Setup keyword arguments.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        A Project instance with setted context.
    """
    if name is not None:
        return get_project(name=name, local=local, config=config, setup_kwargs=setup_kwargs)
    if filename is not None:
        return import_project(filename, local=local, config=config, setup_kwargs=setup_kwargs)
    raise EntityError("Either name or filename must be provided.")


def get_or_create_project(
    name: str,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Get or create a project.

    Parameters
    ----------
    name : str
        Project name.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore env configuration.
    setup_kwargs : dict
        Setup keyword arguments.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        A Project instance.
    """
    try:
        return get_project(name, local, config=config, setup_kwargs=setup_kwargs, **kwargs)
    except BackendError:
        return new_project(name, local=local, config=config, setup_kwargs=setup_kwargs, **kwargs)


def new_project(
    name: str,
    description: str | None = None,
    source: str | None = None,
    labels: list[str] | None = None,
    local: bool = False,
    config: dict | None = None,
    context: str | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Create project.

    Parameters
    ----------
    name : str
        Name that identifies the object.
    kind : str
        Kind of the object.
    uuid : str
        ID of the object in form of UUID.
    description : str
        Description of the object.
    source : str
        Remote git source for object.
    labels : list[str]
        List of labels.
    local : bool
        Flag to determine if object will be exported to backend.
    config : dict
        DHCore env configuration.
    context : str
        The context of the project.
    setup_kwargs : dict
        Setup keyword arguments.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Project
        Project object.
    """
    build_client(local, config)
    obj = create_project(
        name=name,
        kind="project",
        description=description,
        source=source,
        labels=labels,
        local=local,
        context=context,
        **kwargs,
    )
    obj.save()
    return _setup_project(obj, setup_kwargs)


def get_project(
    name: str,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Retrieves project details from backend.

    Parameters
    ----------
    name : str
        The Project name.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore env configuration.
    setup_kwargs : dict
        Setup keyword arguments.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Project
        Object instance.
    """
    build_client(local, config)
    api = api_base_read("projects", name)
    client = get_client(local)
    obj = client.read_object(api, **kwargs)
    obj["local"] = local
    project = create_project_from_dict(obj)
    return _setup_project(project, setup_kwargs)


def import_project(
    file: str,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
) -> Project:
    """
    Import an Project object from a file using the specified file path.

    Parameters
    ----------
    file : str
        Path to the file.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore env configuration.

    Returns
    -------
    Project
        Object instance.
    """
    build_client(local, config)
    obj: dict = read_yaml(file)
    obj["local"] = local
    project = create_project_from_dict(obj)
    return _setup_project(project, setup_kwargs)


def delete_project(
    name: str, cascade: bool = True, clean_context: bool = True, local: bool = False, **kwargs
) -> list[dict]:
    """
    Delete a project.

    Parameters
    ----------
    name : str
        Project name.
    cascade : bool
        Flag to determine if delete is cascading.
    clean_context : bool
        Flag to determine if context will be deleted. If a context is deleted,
        all its objects are unreacheable.
    local : bool
        Flag to determine if backend is local.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    client = get_client(local)
    api = api_base_delete("projects", name)
    params = kwargs.get("params", {})
    if params is None or not params:
        kwargs["params"] = {}
    kwargs["params"]["cascade"] = str(cascade).lower()
    response = client.delete_object(api, **kwargs)
    if clean_context:
        delete_context(name)
    return response


def update_project(entity: Project, local: bool = False, **kwargs) -> dict:
    """
    Update object in backend.

    Parameters
    ----------
    entity : Project
        The object to update.
    local : bool
        Flag to determine if backend is local.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_update("projects", entity.name)
    client = get_client(local)
    return client.update_object(api, entity.to_dict(), **kwargs)


def _setup_project(project: Project, setup_kwargs: dict = None) -> Project:
    """
    Search for setup_project.py file and launch setup
    hanlder as project hook.

    Parameters
    ----------
    project : Project
        The project to scafold.
    setup_kwargs : dict
        Arguments to pass to setup handler.

    Returns
    -------
    Project
        Set up project.
    """
    if setup_kwargs is None:
        setup_kwargs = {}
    check_pth = Path(project.spec.context, ".CHECK")
    setup_pth = Path(project.spec.context, "setup_project.py")
    if setup_pth.exists() and not check_pth.exists():
        spec = imputil.spec_from_file_location("setup_project", setup_pth)
        mod = imputil.module_from_spec(spec)
        spec.loader.exec_module(mod)
        handler = getattr(mod, "setup")
        project = handler(project, **setup_kwargs)
        check_pth.touch()
    return project



"""
Entity metadata module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj


class Metadata(ModelObj):
    """
    A class representing the metadata of an entity.
    Metadata is a collection of information about an entity thought
    to be modifiable by the user. The information contained in the
    metadata can be discordant with the actual state of the entity,
    for example the name of the entity in the database.
    """

    def __init__(
        self,
        name: str | None = None,
        source: str | None = None,
        labels: list[str] | None = None,
        embedded: bool = True,
        created: str | None = str,
        updated: str | None = str,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        name : str
            Name the object.
        source : str
            (Remote GIT) Source of the entity.
        labels : list[str]
            A list of labels to associate with the entity.
        embedded : bool
            Whether the entity specifications are embedded into a project.
        created : str
            Created date.
        updated : str
            Updated date.
        """
        self.name = name
        self.source = source
        self.labels = labels
        self.embedded = embedded
        self.created = created
        self.updated = updated

    @classmethod
    def from_dict(cls, obj: dict) -> "Metadata":
        """
        Return entity metadata object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity metadata.

        Returns
        -------
        Metadata
            An entity metadata object.
        """
        return cls(**obj)


class MetadataRegistry(dict):
    """
    A class representing the registry of entity metadata.
    """

    def register(self, kind: str, module: str, metadata_class: str) -> None:
        """
        Register an entity metadata.

        Parameters
        ----------
        kind : str
            The kind of the entity.
        module : str
            The module name of the entity metadata.
        metadata_class : str
            The class name of the entity metadata.

        Returns
        -------
        None
        """
        self[kind] = {"module": module, "metadata_class": metadata_class}



"""
Status class module.
"""
from __future__ import annotations

from enum import Enum

from digitalhub_core.entities._base.base import ModelObj


class State(Enum):
    """
    State enumeration.
    """

    BUILT = "BUILT"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    CREATED = "CREATED"
    CREATING = "CREATING"
    DELETED = "DELETED"
    ERROR = "ERROR"
    FSM_ERROR = "FSM_ERROR"
    IDLE = "IDLE"
    NONE = "NONE"
    ONLINE = "ONLINE"
    PENDING = "PENDING"
    READY = "READY"
    RUN_ERROR = "RUN_ERROR"
    RUNNING = "RUNNING"
    STOP = "STOP"
    STOPPED = "STOPPED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"


class Status(ModelObj):
    """
    Base Status class.
    The status class contains information about the state of an entity,
    for example, the state of a RUNNING run, and eventual error message.
    """

    def __init__(self, state: str, message: str | None = None) -> None:
        """
        Constructor.

        Parameters
        ----------
        state : str
            The state of the entity.
        message : str
            Error message.
        """
        self.state = state
        self.message = message

    @classmethod
    def from_dict(cls, obj: dict) -> "Status":
        """
        Return entity status object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity status.

        Returns
        -------
        EntityStatus
            An entity status object.
        """
        return cls(**obj)


class StatusRegistry(dict):
    """
    A class representing the registry of entity statuss.
    """

    def register(self, kind: str, module: str, status_class: str) -> None:
        """
        Register an entity status.

        Parameters
        ----------
        kind : str
            The kind of the entity.
        module : str
            The module name of the entity status.
        status_class : str
            The class name of the entity status.

        Returns
        -------
        None
        """
        self[kind] = {"module": module, "status_class": status_class}



"""
Import utils module.
"""
from __future__ import annotations

import importlib
import importlib.metadata


def check_module_existence_by_framework(framework: str) -> str:
    """
    Try to import registry from implemented module.

    Parameters
    ----------
    framework : str
        Framework name.

    Returns
    -------
    str
        Valid module name.
    """
    # Cycle over digitalhub layers modules (core, data, ml, ai).
    for layer in ["core", "data", "ml", "ai"]:
        # Try to import module
        module_name = f"digitalhub_{layer}_{framework}"
        try:
            importlib.metadata.distribution(module_name)
            return module_name
        except importlib.metadata.PackageNotFoundError:
            continue
    # If no module is found raise error
    else:
        raise ModuleNotFoundError(f"Module not found in digitalhub layers core, data, ml, ai for framework {framework}")


def check_layer_existence(layer: str) -> None:
    """
    Check if layer exists.

    Parameters
    ----------
    layer : str
        Layer name.

    Returns
    -------
    None
    """
    try:
        importlib.metadata.distribution(layer)
    except importlib.metadata.PackageNotFoundError:
        raise ModuleNotFoundError(f"Layer {layer} not found.")


def import_registry(module_to_import: str, registry: str) -> dict:
    """
    Import registry from implemented module.

    Parameters
    ----------
    module_to_import : str
        Module name.
    registry : str
        Registry name.

    Returns
    -------
    dict
        Registry.
    """
    module = importlib.import_module(module_to_import)
    try:
        return getattr(module, registry)
    except AttributeError:
        raise ModuleNotFoundError(f"Module {module_to_import} has no '{registry}'.")


def import_class(module_to_import: str, class_name: str) -> type:
    """
    Import class from implemented module.

    Parameters
    ----------
    module_to_import : str
        Module name.
    class_name : str
        Class name.

    Returns
    -------
    type
        Class.
    """
    module = importlib.import_module(module_to_import)
    try:
        return getattr(module, class_name)
    except AttributeError:
        raise ModuleNotFoundError(f"Module {module_to_import} has no '{class_name}' class.")



"""
Run base specification module.
"""
from __future__ import annotations

import typing
from typing import Union

from digitalhub_core.entities._base.spec import Spec, SpecParams
from digitalhub_core.entities.artifacts.crud import get_artifact_from_key
from digitalhub_core.utils.generic_utils import parse_entity_key

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.entity import Entity


ENTITY_FUNC = {
    "artifacts": get_artifact_from_key,
}


class RunSpec(Spec):
    """Run specification."""

    def __init__(
        self,
        task: str,
        inputs: dict | None = None,
        outputs: dict | None = None,
        parameters: dict | None = None,
        values: list | None = None,
        local_execution: bool = False,
    ) -> None:
        """
        Constructor.
        """
        self.task = task
        self.inputs = inputs if inputs is not None else []
        self.outputs = outputs if outputs is not None else []
        self.parameters = parameters if parameters is not None else {}
        self.values = values if values is not None else []
        self.local_execution = local_execution

    def get_inputs(self, as_dict: bool = False) -> list[dict[str, Entity]]:
        """
        Get inputs.

        Returns
        -------
        list[dict[str, Entity]]
            The inputs.
        """
        inputs = []
        for i in self.inputs:
            for k, v in i.items():
                parameter, entity_type, _ = self._parse_input_parameter(k)

                # TODO: Check if entity exists, otherwise create it

                if entity_type is None:
                    # Get entity by type from entity key
                    if v.startswith("store://"):
                        _, entity_type, _, _, _ = parse_entity_key(v)
                        entity = ENTITY_FUNC[entity_type](v)
                        if as_dict:
                            entity = entity.to_dict()
                        inputs.append({parameter: entity})
                    else:
                        raise ValueError(f"Invalid entity key: {v}")

                # TODO: Create new entity
                else:
                    raise NotImplementedError

        return inputs

    @staticmethod
    def _parse_input_parameter(key: str) -> tuple:
        """
        Parse input parameter.

        Parameters
        ----------
        key : str
            Input key.

        Returns
        -------
        tuple
            The parsed parameter.
        """
        # Get parameter name
        splitted = key.split(":")
        parameter = splitted[0]

        try:
            splitted = splitted[1].split(".")
        except IndexError:
            return parameter, None, None

        # Get entity type
        if len(splitted) == 1:
            return parameter, splitted[0], None

        # Get entity kind
        return parameter, splitted[0], splitted[1]


class RunParams(SpecParams):
    """
    Run parameters.
    """

    task: str = None
    """The task string associated with the run."""

    inputs: list[dict[str, Union[str, dict]]] = None
    """Run inputs."""

    outputs: list[dict[str, Union[str, dict]]] = None
    """Run outputs."""

    parameters: dict = None
    """Parameters to be used in the run."""

    values: list = None
    """Values to be used in the run."""

    local_execution: bool = False
    """Flag to indicate if the run will be executed locally."""



"""
Task specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec, SpecParams
from digitalhub_core.entities.tasks.models import Affinity, Env, Label, NodeSelector, Resource, Toleration, Volume


class TaskSpec(Spec):
    """Task specification."""

    def __init__(
        self,
        function: str,
        node_selector: list[NodeSelector] | None = None,
        volumes: list[Volume] | None = None,
        resources: list[Resource] | None = None,
        labels: list[Label] | None = None,
        affinity: Affinity | None = None,
        tolerations: list[Toleration] | None = None,
        env: list[Env] | None = None,
        secrets: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        function : str
            The function string of the task.
        node_selector : list[NodeSelector]
            The node selector of the task.
        volumes : list[Volume]
            The volumes of the task.
        resources : list[Resource]
            Kubernetes resources for the task.
        affinity : Affinity
            The affinity of the task.
        tolerations : list[Toleration]
            The tolerations of the task.
        labels : list[Label]
            The labels of the task.
        env : list[Env]
            The env variables of the task.
        secrets : list[str]
            The secrets of the task.
        """
        self.function = function
        self.node_selector = node_selector
        self.volumes = volumes
        self.resources = resources
        self.labels = labels
        self.affinity = affinity
        self.tolerations = tolerations
        self.env = env
        self.secrets = secrets

        self._any_setter(**kwargs)


class TaskParams(SpecParams):
    """
    Base task model.
    """

    function: str
    """Function string."""

    node_selector: list[NodeSelector] = None
    """Node selector."""

    volumes: list[Volume] = None
    """List of volumes."""

    resources: list[Resource] = None
    """Resources restrictions."""

    affinity: Affinity = None
    """Affinity."""

    tolerations: list[Toleration] = None
    """Tolerations."""

    labels: list[Label] = None
    """List of labels."""

    env: list[Env] = None
    """Env variables."""

    secrets: list[str] = None
    """List of secret names."""



"""
RunStatus class module.
"""
from __future__ import annotations

import typing

from digitalhub_core.entities._base.status import Status
from digitalhub_core.entities.artifacts.crud import get_artifact_from_key
from digitalhub_core.utils.generic_utils import parse_entity_key

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.entity import Entity

ENTITY_FUNC = {
    "artifacts": get_artifact_from_key,
}


class RunStatus(Status):
    """
    Status class for run entities.
    """

    def __init__(
        self,
        state: str,
        message: str | None = None,
        outputs: list | None = None,
        values: list | None = None,
        results: dict | None = None,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(state, message)
        self.outputs = outputs
        self.values = values
        self.results = results

    def get_results(self) -> dict:
        """
        Get results.

        Returns
        -------
        dict
            The results.
        """
        return self.results if self.results is not None else {}

    def get_outputs(self, as_key: bool = True, as_dict: bool = False) -> list[dict[str, str | dict | Entity]]:
        """
        Get outputs.

        Returns
        -------
        list[dict[str, str | dict | Entity]]
            The outputs.
        """
        outputs = []
        if self.outputs is None:
            return outputs
        for i in self.outputs:
            for k, v in i.items():
                _, entity_type, _, _, _ = parse_entity_key(v)
                if v.startswith("store://"):
                    entity = ENTITY_FUNC[entity_type](v)
                    if as_key:
                        entity = entity.key
                    if as_dict:
                        entity = entity.to_dict()
                    outputs.append({k: entity})
                else:
                    raise ValueError(f"Invalid entity key: {v}")
        return outputs

    def get_values(self) -> list:
        """
        Get values.

        Returns
        -------
        list
            The values.
        """
        return self.values if self.values is not None else []



"""
General utilities module.
"""
from __future__ import annotations

import base64
import os
from datetime import datetime
from hashlib import sha1
from mimetypes import guess_type
from pathlib import Path
from uuid import uuid4

from digitalhub_core.utils.io_utils import read_text


def build_uuid(uuid: str | None = None) -> str:
    """
    Create a uuid if not given

    Parameters
    ----------
    uuid : str
        ID of the object in form of UUID. Optional.

    Returns
    -------
    str
        The uuid.
    """
    if uuid is not None:
        return uuid
    return str(uuid4())


def get_timestamp() -> str:
    """
    Get the current timestamp timezoned.

    Returns
    -------
    str
        The current timestamp.
    """
    return datetime.now().astimezone().isoformat()


def decode_string(string: str) -> str:
    """
    Decode a string from base64.

    Parameters
    ----------
    string : str
        The string to decode.

    Returns
    -------
    str
        The string decoded from base64.
    """
    return base64.b64decode(string).decode()


def encode_string(string: str) -> str:
    """
    Encode a string in base64.

    Parameters
    ----------
    string : str
        The string to encode.

    Returns
    -------
    str
        The string encoded in base64.
    """
    return base64.b64encode(string.encode()).decode()


def encode_source(path: str) -> str:
    """
    Read a file and encode in base64 the content.

    Parameters
    ----------
    path : str
        The file path to read.

    Returns
    -------
    str
        The file content encoded in base64.
    """
    return encode_string(read_text(path))


def set_dhub_env(
    endpoint: str | None = None,
    user: str | None = None,
    password: str | None = None,
    token: str | None = None,
) -> None:
    """
    Function to set environment variables for DHub Core config.

    Parameters
    ----------
    endpoint : str
        The endpoint of DHub Core.
    user : str
        The user of DHub Core.
    password : str
        The password of DHub Core.
    token : str
        The token of DHub Core.

    Returns
    -------
    None
    """
    if endpoint is not None:
        os.environ["DIGITALHUB_CORE_ENDPOINT"] = endpoint
    if user is not None:
        os.environ["DIGITALHUB_CORE_USER"] = user
    if password is not None:
        os.environ["DIGITALHUB_CORE_PASSWORD"] = password
    if token is not None:
        os.environ["DIGITALHUB_CORE_TOKEN"] = token


def parse_entity_key(key: str) -> tuple[str]:
    """
    Parse the entity key.

    Parameters
    ----------
    key : str
        The entity key.

    Returns
    -------
    tuple[str]
        Project, entity type, kind, name and uuid.
    """
    try:
        # Remove "store://" from the key
        key = key.replace("store://", "")
        # Split the key into parts
        parts = key.split("/")
        # The project is the first part
        project = parts[0]
        # The entity type is the second part
        entity_type = parts[1]
        # The kind is the third part
        kind = parts[2]
        # The name and uuid are separated by a colon in the last part
        name, uuid = parts[3].split(":")
        return project, entity_type, kind, name, uuid
    except Exception as e:
        raise ValueError("Invalid key format.") from e


def calculate_blob_hash(data_path: str) -> str:
    """
    Calculate the hash of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    str
        The hash of the file.
    """
    with open(data_path, "rb") as f:
        data = f.read()
        return f"sha1_{sha1(data).hexdigest()}"


def get_file_size(data_path: str) -> int:
    """
    Get the size of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    int
        The size of the file.
    """
    return Path(data_path).stat().st_size


def get_file_mime_type(data_path: str) -> str:
    """
    Get the mime type of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    str
        The mime type of the file.
    """
    return guess_type(data_path)[0]


def get_file_extension(data_path: str) -> str:
    """
    Get the extension of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    str
        The extension of the file.
    """
    return Path(data_path).suffix[1:]



from __future__ import annotations

from digitalhub_core.entities._base.metadata import MetadataRegistry
from digitalhub_core.entities._base.spec import SpecRegistry
from digitalhub_core.entities._base.status import StatusRegistry

metadata_registry = MetadataRegistry()
spec_registry = SpecRegistry()
status_registry = StatusRegistry()

for i in [
    "artifact",
    "project",
    "secret",
    "service",
    "workflow",
]:
    root = f"digitalhub_core.entities.{i}s"
    metadata_registry.register(i, f"{root}.metadata", f"{i.title()}Metadata")
    spec_registry.register(i, f"{root}.spec", f"{i.title()}Spec", f"{i.title()}Params")
    status_registry.register(i, f"{root}.status", f"{i.title()}Status")
